# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest2
import datetime
from stalker import (defaults, Studio, WorkingHours, Project, StatusList,
                     Status, Repository, Asset, Type, User, Shot, Department,
                     Task, TaskJugglerScheduler)
from stalker import db
from stalker.db import DBSession

# for IDEs to help them in code compilation
if False:
    from sqlalchemy.orm import Session

    assert isinstance(DBSession, Session)


class StudioTester(unittest2.TestCase):
    """tests the stalker.models.studio.Studio class
    """

    @classmethod
    def setUpClass(cls):
        """setup the test
        """
        DBSession.configure(extension=None)

    @classmethod
    def tearDownClass(cls):
        """cleanup the test
        """
        DBSession.configure(extension=None)

    def setUp(self):
        """setup the test
        """
        db.setup({
            "sqlalchemy.url": "sqlite:///:memory:",
            "sqlalchemy.echo": False,
        })
        db.init()

        self.test_user1 = User(
            name='User 1',
            login='user1',
            email='user1@users.com',
            password='password'
        )
        DBSession.add(self.test_user1)

        self.test_user2 = User(
            name='User 2',
            login='user2',
            email='user2@users.com',
            password='password'
        )
        DBSession.add(self.test_user2)

        self.test_user3 = User(
            name='User 3',
            login='user3',
            email='user3@users.com',
            password='password'
        )
        DBSession.add(self.test_user3)

        self.test_department1 = Department(
            name='Test Department 1'
        )
        DBSession.add(self.test_department1)

        self.test_department2 = Department(
            name='Test Department 2'
        )
        DBSession.add(self.test_department2)

        self.test_status1 = Status(
            name='Status 1',
            code='ST1'
        )
        DBSession.add(self.test_status1)

        self.test_status2 = Status(
            name='Status 2',
            code='ST2'
        )
        DBSession.add(self.test_status2)

        self.test_project_status_list1 = StatusList(
            name='ProjectStatuses',
            statuses=[self.test_status1, self.test_status2],
            target_entity_type=Project
        )
        DBSession.add(self.test_project_status_list1)

        self.test_repo = Repository(
            name='Test Repository',
            windows_path='T:/',
            linux_path='/mnt/T/',
            osx_path='/Volumes/T/'
        )
        DBSession.add(self.test_repo)

        # create a couple of projects
        self.test_project1 = Project(
            name='Test Project 1',
            code='TP1',
            status_list=self.test_project_status_list1,
            repository=self.test_repo
        )
        DBSession.add(self.test_project1)

        self.test_project2 = Project(
            name='Test Project 2',
            code='TP2',
            status_list=self.test_project_status_list1,
            repository=self.test_repo
        )
        DBSession.add(self.test_project2)

        # an inactive project
        self.test_project3 = Project(
            name='Test Project 3',
            code='TP3',
            status_list=self.test_project_status_list1,
            repository=self.test_repo
        )
        self.test_project3.active = False
        DBSession.add(self.test_project3)

        # create assets and shots
        self.test_asset_type = Type(
            name='Character',
            code='Char',
            target_entity_type=Asset
        )
        DBSession.add(self.test_asset_type)

        self.test_asset_status_list = StatusList(
            name='AssetStatuses',
            statuses=[self.test_status1, self.test_status2],
            target_entity_type=Asset,
        )
        DBSession.add(self.test_asset_status_list)

        self.test_asset1 = Asset(
            name='Test Asset 1',
            code='TA1',
            project=self.test_project1,
            status_list=self.test_asset_status_list,
            type=self.test_asset_type
        )
        DBSession.add(self.test_asset1)

        self.test_asset2 = Asset(
            name='Test Asset 2',
            code='TA2',
            project=self.test_project2,
            status_list=self.test_asset_status_list,
            type=self.test_asset_type
        )
        DBSession.add(self.test_asset2)

        # shots
        self.test_shot_status_list = StatusList(
            name='ShotStatuses',
            statuses=[self.test_status1, self.test_status2],
            target_entity_type=Shot
        )
        DBSession.add(self.test_shot_status_list)

        # for project 1
        self.test_shot1 = Shot(
            code='shot1',
            project=self.test_project1,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot1)

        self.test_shot2 = Shot(
            code='shot2',
            project=self.test_project1,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot2)

        # for project 2
        self.test_shot3 = Shot(
            code='shot3',
            project=self.test_project2,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot3)

        self.test_shot4 = Shot(
            code='shot4',
            project=self.test_project2,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot4)

        # for project 3
        self.test_shot5 = Shot(
            code='shot5',
            project=self.test_project3,
            status_list=self.test_shot_status_list
        )
        DBSession.add(self.test_shot5)

        #########################################################3
        # tasks for projects
        self.test_task_statuses = StatusList(
            name='Task Statuses',
            statuses=[self.test_status1, self.test_status2],
            target_entity_type=Task,
        )
        DBSession.add(self.test_task_statuses)

        self.test_task1 = Task(
            name='Project Planing',
            project=self.test_project1,
            resources=[self.test_user1],
            status_list=self.test_task_statuses,
            schedule_timing=10,
            schedule_unit='d'
        )
        DBSession.add(self.test_task1)

        self.test_task2 = Task(
            name='Project Planing',
            project=self.test_project2,
            resources=[self.test_user1],
            status_list=self.test_task_statuses,
            schedule_timing=10,
            schedule_unit='d'
        )
        DBSession.add(self.test_task2)

        self.test_task3 = Task(
            name='Project Planing',
            project=self.test_project3,
            resources=[self.test_user1],
            status_list=self.test_task_statuses,
            schedule_timing=5,
            schedule_unit='d'
        )
        DBSession.add(self.test_task3)

        # for shots

        # Shot 1
        self.test_task4 = Task(
            name='Match Move',
            parent=self.test_shot1,
            resources=[self.test_user1],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task4)

        self.test_task5 = Task(
            name='FX',
            parent=self.test_shot1,
            resources=[self.test_user2],
            depends=[self.test_task4],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task5)

        self.test_task6 = Task(
            name='Lighting',
            parent=self.test_shot1,
            resources=[self.test_user2],
            depends=[self.test_task4, self.test_task5],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task6)

        self.test_task7 = Task(
            name='Comp',
            parent=self.test_shot1,
            resources=[self.test_user2],
            depends=[self.test_task6],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task7)

        # Shot 2
        self.test_task8 = Task(
            name='Match Move',
            parent=self.test_shot2,
            resources=[self.test_user3],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task8)

        self.test_task9 = Task(
            name='FX',
            parent=self.test_shot2,
            resources=[self.test_user3],
            depends=[self.test_task8],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task9)

        self.test_task10 = Task(
            name='Lighting',
            parent=self.test_shot2,
            resources=[self.test_user2],
            depends=[self.test_task8, self.test_task9],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task10)

        self.test_task11 = Task(
            name='Comp',
            parent=self.test_shot2,
            resources=[self.test_user2],
            depends=[self.test_task10],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task11)

        # Shot 3
        self.test_task12 = Task(
            name='Match Move',
            parent=self.test_shot3,
            resources=[self.test_user1],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task12)

        self.test_task13 = Task(
            name='FX',
            parent=self.test_shot3,
            resources=[self.test_user1],
            depends=[self.test_task12],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task13)

        self.test_task14 = Task(
            name='Lighting',
            parent=self.test_shot3,
            resources=[self.test_user1],
            depends=[self.test_task12, self.test_task13],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task14)

        self.test_task15 = Task(
            name='Comp',
            parent=self.test_shot3,
            resources=[self.test_user1],
            depends=[self.test_task14],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task15)

        # Shot 4
        self.test_task16 = Task(
            name='Match Move',
            parent=self.test_shot4,
            resources=[self.test_user2],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task16)

        self.test_task17 = Task(
            name='FX',
            parent=self.test_shot4,
            resources=[self.test_user2],
            depends=[self.test_task16],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task17)

        self.test_task18 = Task(
            name='Lighting',
            parent=self.test_shot4,
            resources=[self.test_user2],
            depends=[self.test_task16, self.test_task17],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task18)

        self.test_task19 = Task(
            name='Comp',
            parent=self.test_shot4,
            resources=[self.test_user2],
            depends=[self.test_task18],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task19)

        # Shot 5
        self.test_task20 = Task(
            name='Match Move',
            parent=self.test_shot5,
            resources=[self.test_user3],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task20)

        self.test_task21 = Task(
            name='FX',
            parent=self.test_shot5,
            resources=[self.test_user3],
            depends=[self.test_task20],
            schedule_timing=2,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task21)

        self.test_task22 = Task(
            name='Lighting',
            parent=self.test_shot5,
            resources=[self.test_user3],
            depends=[self.test_task20, self.test_task21],
            schedule_timing=3,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task22)

        self.test_task23 = Task(
            name='Comp',
            parent=self.test_shot5,
            resources=[self.test_user3],
            depends=[self.test_task22],
            schedule_timing=4,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task23)

        ####################################################3
        # For Assets

        # Asset 1
        self.test_task24 = Task(
            name='Design',
            parent=self.test_asset1,
            resources=[self.test_user1],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task24)

        self.test_task25 = Task(
            name='Model',
            parent=self.test_asset1,
            depends=[self.test_task24],
            resources=[self.test_user1],
            schedule_timing=15,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task25)

        self.test_task26 = Task(
            name='LookDev',
            parent=self.test_asset1,
            depends=[self.test_task25],
            resources=[self.test_user1],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task26)

        self.test_task27 = Task(
            name='Rig',
            parent=self.test_asset1,
            depends=[self.test_task25],
            resources=[self.test_user1],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task27)

        # Asset 2
        self.test_task28 = Task(
            name='Design',
            parent=self.test_asset2,
            resources=[self.test_user2],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task28)

        self.test_task29 = Task(
            name='Model',
            parent=self.test_asset2,
            depends=[self.test_task28],
            resources=[self.test_user2],
            schedule_timing=15,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task29)

        self.test_task30 = Task(
            name='LookDev',
            parent=self.test_asset2,
            depends=[self.test_task29],
            resources=[self.test_user2],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task30)

        self.test_task31 = Task(
            name='Rig',
            parent=self.test_asset2,
            depends=[self.test_task29],
            resources=[self.test_user2],
            schedule_timing=10,
            schedule_unit='d',
            status_list=self.test_task_statuses
        )
        DBSession.add(self.test_task31)

        # TODO: Add Milestones

        self.kwargs = dict(
            name='Studio',
            daily_working_hours=8,
        )

        self.test_studio = Studio(**self.kwargs)
        DBSession.add(self.test_studio)
        DBSession.commit()

    def test_daily_working_hours_argument_is_skipped(self):
        """testing if the daily_working_hours attribute will be equal to the
        default settings when the daily_working_hours argument is skipped
        """
        try:
            self.kwargs.pop('daily_working_hours')
        except KeyError:
            pass
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.daily_working_hours,
                         defaults.daily_working_hours)

    def test_daily_working_hours_argument_is_None(self):
        """testing if the daily_working_hours attribute will be equal to the
        default settings value when the daily_working_hours argument is None
        """
        self.kwargs['daily_working_hours'] = None
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.daily_working_hours,
                         defaults.daily_working_hours)

    def test_daily_working_hours_attribute_is_None(self):
        """testing if the daily_working_hours attribute will be equal to the
        default settings value when it is set to None
        """
        self.test_studio.daily_working_hours = None
        self.assertEqual(self.test_studio.daily_working_hours,
                         defaults.daily_working_hours)

    def test_daily_working_hours_argument_is_not_integer(self):
        """testing if a TypeError will be raised when the daily_working_hours
        argument is not an integer
        """
        self.kwargs['daily_working_hours'] = 'not an integer'
        self.assertRaises(TypeError, Studio, **self.kwargs)

    def test_daily_working_hours_attribute_is_not_an_integer(self):
        """testing if a TypeError will be raised when the daily_working hours
        attribute is set to a value other than an integer
        """
        self.assertRaises(TypeError, setattr, self.test_studio,
                          'daily_working_hours', 'not an intger')

    def test_daily_working_hours_argument_is_working_fine(self):
        """testing if the daily working hours argument value is correctly
        passed to daily_working_hours attribute
        """
        self.kwargs['daily_working_hours'] = 12
        new_project = Studio(**self.kwargs)
        self.assertEqual(new_project.daily_working_hours, 12)

    def test_daily_working_hours_attribute_is_working_properly(self):
        """testing if the daily_working_hours attribute is working properly
        """
        self.test_studio.daily_working_hours = 23
        self.assertEqual(self.test_studio.daily_working_hours, 23)

    def test_working_hours_argument_is_skipped(self):
        """testing if the default working hours will be used when the
        working_hours argument is skipped
        """
        self.kwargs['name'] = 'New Studio'
        try:
            self.kwargs.pop('working_hours') # pop if there are any
        except KeyError:
            pass

        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.working_hours, WorkingHours())

    def test_working_hours_argument_is_None(self):
        """testing if the a WorkingHour instance with default settings will be
        used if the working_hours argument is skipped
        """
        self.kwargs['name'] = 'New Studio'
        self.kwargs['working_hours'] = None
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.working_hours, WorkingHours())

    def test_working_hours_attribute_is_None(self):
        """testing if a WorkingHour instance will be created with the default
        values if the working_hours attribute is set to None
        """
        self.test_studio.working_horus = None
        self.assertEqual(self.test_studio.working_hours, WorkingHours())

    def test_tjp_id_attribute_returns_a_plausible_id(self):
        """testing if the tjp_id is returning something meaningful
        """
        self.test_studio.id = 432
        self.assertEqual(self.test_studio.tjp_id, 'Studio_432')

    def test_projects_attribute_is_read_only(self):
        """testing if the project attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'projects', [self.test_project1])

    def test_projects_attribute_is_working_properly(self):
        """testing if the projects attribute is working properly
        """
        self.assertItemsEqual(
            self.test_studio.projects,
            [self.test_project1, self.test_project2, self.test_project3]
        )

    def test_active_projects_attribute_is_read_only(self):
        """testing if the active_projects attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'active_projects', [self.test_project1])

    def test_active_projects_attribute_is_working_properly(self):
        """testing if the active_projects attribute is working properly
        """
        self.assertItemsEqual(
            self.test_studio.active_projects,
            [self.test_project1, self.test_project2]
        )

    def test_inactive_projects_attribute_is_read_only(self):
        """testing if the inactive_projects attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'inactive_projects', [self.test_project1])

    def test_inactive_projects_attribute_is_working_properly(self):
        """testing if the inactive_projects attribute is working properly
        """
        self.assertItemsEqual(
            self.test_studio.inactive_projects,
            [self.test_project3]
        )

    def test_departments_attribute_is_read_only(self):
        """testing if the departments attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio,
                          'departments', [self.test_project1])

    def test_departments_attribute_is_working_properly(self):
        """testing if the departments attribute is working properly
        """
        # don't forget admins department
        admins_dep = Department.query.filter_by(name='admins').first()
        self.assertItemsEqual(self.test_studio.departments,
                              [self.test_department1, self.test_department2,
                               admins_dep])

    def test_users_attribute_is_read_only(self):
        """testing if the users attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio, 'users',
                          [self.test_project1])

    def test_users_attribute_is_working_properly(self):
        """testing if the users attribute is working properly
        """
        # don't forget the admin
        admin = User.query.filter_by(name='admin').first()

        self.assertItemsEqual(
            self.test_studio.users,
            [admin, self.test_user1, self.test_user2, self.test_user3]
        )

    def test_to_tjp_attribute_is_read_only(self):
        """testing if the to_tjp attribute is a read only attribute
        """
        self.assertRaises(AttributeError, setattr, self.test_studio, 'to_tjp',
                          "some text")

    def test_now_argument_is_skipped(self):
        """testing if the now attribute will use the rounded
        datetime.datetime.now value when the now argument is skipped
        """
        try:
            self.kwargs.pop('now')
        except KeyError:
            pass

        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.now,
                         new_studio.round_time(datetime.datetime.now()))

    def test_now_argument_is_None(self):
        """testing if the now attribute will use the rounded
        datetime.datetime.now value when the now argument is None
        """
        self.kwargs['now'] = None
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.now,
                         new_studio.round_time(datetime.datetime.now()))

    def test_now_attribute_is_None(self):
        """testing if the now attribute will be equal to the rounded value of
        datetime.datetime.now() if it is set to None
        """
        self.test_studio.now = None
        self.assertEqual(
            self.test_studio.now,
            self.test_studio.round_time(datetime.datetime.now())
        )

    def test_now_argument_is_not_a_datetime_instance(self):
        """testing if a TypeError will be raised when the now argument is not
        a datetime.datetime instance
        """
        self.kwargs['now'] = 'not a datetime instance'
        self.assertRaises(TypeError, Studio, **self.kwargs)

    def test_now_attribute_is_set_to_a_value_other_than_datetime_instance(
            self):
        """testing if a TypeError will be raised when the now attribute is set
        to a value other than a datetime.datetime instance
        """
        self.assertRaises(TypeError, setattr, self.test_studio, 'now',
                          'not a datetime instance')

    def test_now_argument_is_working_properly(self):
        """testing if the now argument value is passed to the now attribute
        properly 
        """
        self.kwargs['now'] = datetime.datetime(2013, 4, 15, 21, 9)
        expected_now = datetime.datetime(2013, 4, 15, 21, 0)
        new_studio = Studio(**self.kwargs)
        self.assertEqual(new_studio.now, expected_now)

    def test_now_attribute_is_working_properly(self):
        """testing if the now attribute is working properly
        """
        self.test_studio.now = datetime.datetime(2013, 4, 15, 21, 11)
        expected_now = datetime.datetime(2013, 4, 15, 21, 0)
        self.assertEqual(self.test_studio.now, expected_now)

    def test_to_tjp_attribute_is_working_properly(self):
        """testing if the to_tjp attribute is working properly
        """
        self.test_studio.start = datetime.datetime(2013, 4, 15, 17, 40)
        self.test_studio.end = datetime.datetime(2013, 6, 30, 17, 40)
        now = self.test_studio.round_time(datetime.datetime.now())
        self.test_studio.working_hours[0] = [[540, 1080]]
        self.test_studio.working_hours[1] = [[540, 1080]]
        self.test_studio.working_hours[2] = [[540, 1080]]
        self.test_studio.working_hours[3] = [[540, 1080]]
        self.test_studio.working_hours[4] = [[540, 1080]]
        self.test_studio.working_hours[5] = [[540, 720]]
        self.test_studio.working_hours[6] = []

        from jinja2 import Template

        expected_tjp_template = Template("""project Studio_{{ studio.id }} "Studio" 2013-04-15 - 2013-06-30 {
            timingresolution 60min
            now {{ studio.now.strftime('%Y-%m-%d-%H:%M') }}
            dailyworkinghours 8
            weekstartsmonday
            workinghours mon 09:00 - 18:00
            workinghours tue 09:00 - 18:00
            workinghours wed 09:00 - 18:00
            workinghours thu 09:00 - 18:00
            workinghours fri 09:00 - 18:00
            workinghours sat 09:00 - 12:00
            workinghours sun off
            timeformat "%Y-%m-%d"
            scenario plan "Plan"
            trackingscenario plan
        }
        """)

        expected_tjp = expected_tjp_template.render({
            'studio': self.test_studio
        })
        #print '-----------------------------------'
        #print expected_tjp
        #print '-----------------------------------'
        #print self.test_studio.to_tjp
        #print '-----------------------------------'
        self.assertEqual(self.test_studio.to_tjp, expected_tjp)

    def test_scheduler_attribute_can_be_set_to_None(self):
        """testing if the scheduler attribute can be set to None
        """
        self.test_studio.scheduler = None

    def test_scheduler_attribute_accepts_Scheduler_instances_only(self):
        """testing if a TypeError will be raised when the scheduler attribute
        is set to a value which is not a scheduler instance
        """
        self.assertRaises(TypeError, setattr, self.test_studio, 'scheduler',
                          'not a Scheduler instance')

    def test_scheduler_attribute_is_working_properly(self):
        """testing if the scheduler attribute is working properly
        """
        tj_s = TaskJugglerScheduler()
        self.test_studio.scheduler = tj_s
        self.assertEqual(self.test_studio.scheduler, tj_s)

    def test_schedule_will_not_work_without_a_scheduler(self):
        """testing if a RuntimeError will be raised when the scheduler
        attribute is not set to a Scheduler instance and schedule is called
        """
        self.test_studio.scheduler = None
        self.assertRaises(RuntimeError, self.test_studio.schedule)

    def test_schedule_will_schedule_the_tasks_with_the_given_scheduler(self):
        """testing if the schedule method will schedule the tasks with the
        given scheduler
        """
        tj_scheduler = TaskJugglerScheduler()
        self.test_studio.now = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.start = datetime.datetime(2013, 4, 15, 22, 56)
        self.test_studio.end = datetime.datetime(2013, 7, 30, 0, 0)

        self.test_studio.scheduler = tj_scheduler
        self.test_studio.schedule()

        # now check the timings of the tasks are all adjusted

        # Projects

        # Project 1
        self.assertEqual(
            self.test_project1.computed_start,
            datetime.datetime(2013, 4, 16, 9, 0)
        )

        self.assertEqual(
            self.test_project1.computed_end,
            datetime.datetime(2013, 7, 19, 13, 0)
        )

        # Project 2
        self.assertEqual(
            self.test_project2.computed_start,
            datetime.datetime(2013, 4, 16, 9, 0)
        )

        self.assertEqual(
            self.test_project1.computed_end,
            datetime.datetime(2013, 7, 19, 13, 0)
        )

        # Project 3
        self.assertIsNone(self.test_project3.computed_start)
        self.assertIsNone(self.test_project3.computed_end)

        # self.test_asset1
        self.assertEqual(
            self.test_asset1.computed_start,
            datetime.datetime(2013, 04, 16, 9, 0)
        )
        self.assertEqual(
            self.test_asset1.computed_end,
            datetime.datetime(2013, 06, 12, 16, 00)
        )

        # self.test_asset2
        self.assertEqual(
            self.test_asset2.computed_start,
            datetime.datetime(2013, 04, 16, 9, 0)
        )
        self.assertEqual(
            self.test_asset2.computed_end,
            datetime.datetime(2013, 06, 12, 16, 00)
        )

        # self.test_shot1
        self.assertEqual(
            self.test_shot1.computed_start,
            datetime.datetime(2013, 07, 8, 14, 0)
        )
        self.assertEqual(
            self.test_shot1.computed_end,
            datetime.datetime(2013, 07, 19, 13, 00)
        )

        # self.test_shot2
        self.assertEqual(
            self.test_shot2.computed_start,
            datetime.datetime(2013, 04, 16, 9, 0)
        )
        self.assertEqual(
            self.test_shot2.computed_end,
            datetime.datetime(2013, 06, 27, 13, 00)
        )

        # self.test_shot3
        self.assertEqual(
            self.test_shot3.computed_start,
            datetime.datetime(2013, 05, 16, 11, 00)
        )
        self.assertEqual(
            self.test_shot3.computed_end,
            datetime.datetime(2013, 07, 22, 12, 00)
        )

        # self.test_shot4
        self.assertEqual(
            self.test_shot4.computed_start,
            datetime.datetime(2013, 05, 16, 11, 00)
        )
        self.assertEqual(
            self.test_shot4.computed_end,
            datetime.datetime(2013, 07, 02, 18, 00)
        )

        # self.test_shot5
        self.assertIsNone(self.test_shot5.computed_start)
        self.assertIsNone(self.test_shot5.computed_end)

        # self.test_task1
        self.assertEqual(
            self.test_task1.computed_start,
            datetime.datetime(2013, 06, 12, 16, 00)
        )
        self.assertEqual(
            self.test_task1.computed_end,
            datetime.datetime(2013, 06, 25, 15, 00)
        )

        # self.test_task2
        self.assertEqual(
            self.test_task2.computed_start,
            datetime.datetime(2013, 06, 25, 15, 00)
        )
        self.assertEqual(
            self.test_task2.computed_end,
            datetime.datetime(2013, 07, 8, 14, 0)
        )

        # self.test_task3
        self.assertIsNone(self.test_task3.computed_start)
        self.assertIsNone(self.test_task3.computed_end)

        # self.test_task4
        self.assertEqual(
            self.test_task4.computed_start,
            datetime.datetime(2013, 07, 8, 14, 0)
        )
        self.assertEqual(
            self.test_task4.computed_end,
            datetime.datetime(2013, 07, 10, 12, 00)
        )

        # self.test_task5
        self.assertEqual(
            self.test_task5.computed_start,
            datetime.datetime(2013, 07, 10, 12, 00)
        )
        self.assertEqual(
            self.test_task5.computed_end,
            datetime.datetime(2013, 07, 12, 10, 00)
        )

        # self.test_task6
        self.assertEqual(
            self.test_task6.computed_start,
            datetime.datetime(2013, 07, 12, 10, 00)
        )
        self.assertEqual(
            self.test_task6.computed_end,
            datetime.datetime(2013, 07, 16, 16, 00)
        )

        # self.test_task7
        self.assertEqual(
            self.test_task7.computed_start,
            datetime.datetime(2013, 07, 16, 16, 00)
        )
        self.assertEqual(
            self.test_task7.computed_end,
            datetime.datetime(2013, 07, 19, 13, 00)
        )

        # self.test_task8
        self.assertEqual(
            self.test_task8.computed_start,
            datetime.datetime(2013, 04, 16, 9, 0)
        )
        self.assertEqual(
            self.test_task8.computed_end,
            datetime.datetime(2013, 04, 17, 16, 00)
        )

        # self.test_task9
        self.assertEqual(
            self.test_task9.computed_start,
            datetime.datetime(2013, 04, 17, 16, 00)
        )
        self.assertEqual(
            self.test_task9.computed_end,
            datetime.datetime(2013, 04, 19, 14, 00)
        )

        # self.test_task10
        self.assertEqual(
            self.test_task10.computed_start,
            datetime.datetime(2013, 06, 14, 14, 00)
        )
        self.assertEqual(
            self.test_task10.computed_end,
            datetime.datetime(2013, 06, 19, 11, 00)
        )

        # self.test_task11
        self.assertEqual(
            self.test_task11.computed_start,
            datetime.datetime(2013, 06, 21, 17, 00)
        )
        self.assertEqual(
            self.test_task11.computed_end,
            datetime.datetime(2013, 06, 27, 13, 00)
        )

        # self.test_task12
        self.assertEqual(
            self.test_task12.computed_start,
            datetime.datetime(2013, 05, 16, 11, 00)
        )
        self.assertEqual(
            self.test_task12.computed_end,
            datetime.datetime(2013, 05, 17, 18, 00)
        )

        # self.test_task13
        self.assertEqual(
            self.test_task13.computed_start,
            datetime.datetime(2013, 07, 10, 12, 00)
        )
        self.assertEqual(
            self.test_task13.computed_end,
            datetime.datetime(2013, 07, 12, 10, 00)
        )

        # self.test_task14
        self.assertEqual(
            self.test_task14.computed_start,
            datetime.datetime(2013, 07, 12, 10, 00)
        )
        self.assertEqual(
            self.test_task14.computed_end,
            datetime.datetime(2013, 07, 16, 16, 00)
        )

        # self.test_task15
        self.assertEqual(
            self.test_task15.computed_start,
            datetime.datetime(2013, 07, 16, 16, 00)
        )
        self.assertEqual(
            self.test_task15.computed_end,
            datetime.datetime(2013, 07, 22, 12, 00)
        )

        # self.test_task16
        self.assertEqual(
            self.test_task16.computed_start,
            datetime.datetime(2013, 05, 16, 11, 00)
        )
        self.assertEqual(
            self.test_task16.computed_end,
            datetime.datetime(2013, 05, 17, 18, 00)
        )

        # self.test_task17
        self.assertEqual(
            self.test_task17.computed_start,
            datetime.datetime(2013, 06, 12, 16, 00)
        )
        self.assertEqual(
            self.test_task17.computed_end,
            datetime.datetime(2013, 06, 14, 14, 00)
        )

        # self.test_task18
        self.assertEqual(
            self.test_task18.computed_start,
            datetime.datetime(2013, 06, 19, 11, 00)
        )
        self.assertEqual(
            self.test_task18.computed_end,
            datetime.datetime(2013, 06, 21, 17, 00)
        )

        # self.test_task19
        self.assertEqual(
            self.test_task19.computed_start,
            datetime.datetime(2013, 06, 27, 13, 00)
        )
        self.assertEqual(
            self.test_task19.computed_end,
            datetime.datetime(2013, 07, 02, 18, 00)
        )

        # self.test_task20
        self.assertIsNone(self.test_task20.computed_start)
        self.assertIsNone(self.test_task20.computed_end)

        # self.test_task21
        self.assertIsNone(self.test_task21.computed_start)
        self.assertIsNone(self.test_task21.computed_end)

        # self.test_task22
        self.assertIsNone(self.test_task22.computed_start)
        self.assertIsNone(self.test_task22.computed_end)

        # self.test_task23
        self.assertIsNone(self.test_task23.computed_start)
        self.assertIsNone(self.test_task23.computed_end)

        # self.test_task24
        self.assertEqual(
            self.test_task24.computed_start,
            datetime.datetime(2013, 04, 16, 9, 0)
        )
        self.assertEqual(
            self.test_task24.computed_end,
            datetime.datetime(2013, 04, 26, 17, 00)
        )

        # self.test_task25
        self.assertEqual(
            self.test_task25.computed_start,
            datetime.datetime(2013, 04, 26, 17, 00)
        )
        self.assertEqual(
            self.test_task25.computed_end,
            datetime.datetime(2013, 05, 16, 11, 00)
        )

        # self.test_task26
        self.assertEqual(
            self.test_task26.computed_start,
            datetime.datetime(2013, 05, 20, 9, 0)
        )
        self.assertEqual(
            self.test_task26.computed_end,
            datetime.datetime(2013, 05, 30, 17, 00)
        )

        # self.test_task27
        self.assertEqual(
            self.test_task27.computed_start,
            datetime.datetime(2013, 05, 30, 17, 00)
        )
        self.assertEqual(
            self.test_task27.computed_end,
            datetime.datetime(2013, 06, 12, 16, 00)
        )

        # self.test_task28
        self.assertEqual(
            self.test_task28.computed_start,
            datetime.datetime(2013, 04, 16, 9, 0)
        )
        self.assertEqual(
            self.test_task28.computed_end,
            datetime.datetime(2013, 04, 26, 17, 00)
        )

        # self.test_task29
        self.assertEqual(
            self.test_task29.computed_start,
            datetime.datetime(2013, 04, 26, 17, 00)
        )
        self.assertEqual(
            self.test_task29.computed_end,
            datetime.datetime(2013, 05, 16, 11, 00)
        )

        # self.test_task30
        self.assertEqual(
            self.test_task30.computed_start,
            datetime.datetime(2013, 05, 20, 9, 0)
        )
        self.assertEqual(
            self.test_task30.computed_end,
            datetime.datetime(2013, 05, 30, 17, 00)
        )

        # self.test_task31
        self.assertEqual(
            self.test_task31.computed_start,
            datetime.datetime(2013, 05, 30, 17, 00)
        )
        self.assertEqual(
            self.test_task31.computed_end,
            datetime.datetime(2013, 06, 12, 16, 00)
        )
    



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import logging


logger = logging.getLogger(__name__)


class EnvironmentBase(object):
    """Connects the environment (the host program) to Stalker.

    In Stalker, an Environment is a host application like Maya, Nuke, Houdini
    etc.

    Generally a GUI for the end user is given an environment which helps the
    QtGui to be able to open, save, import or export a Version without
    knowing the details of the environment.

    The environment object supplies **methods** like ``open``, ``save``,
    ``export``,  ``import`` or ``reference``. The main duty of the Environment
    object is to introduce the host application (Maya, Houdini, Nuke, etc.) to
    Stalker and let it to open, save, export, import or reference a file.

    It is the pipeline developers duty to create the environment classes for
    the programs used in the studio by instantiating this class and overriding
    the methods as necessary. You can find good examples in `Anima Tools`_
    which is a Python package developed in `Anima Istanbul`_.

    .. _Anima Tools: https://pypi.python.org/pypi/anima
    .. _Anima Istanbul: http;//www.animaistanbul.com

    Here is a brief example for creating an environment for a generic program::

        from Stalker import EnvironmentBase

        class MyProgramEnv(EnvironmentBase):
            \"""This is a class which will be used by the UI
            \"""

            def open():
                \"""uses the programs own Python API to open a version of an
                asset
                \"""

                # do anything that needs to be done before opening the file
                my_programs_own_python_api.open(filepath=self.version.full_path)

            def save():
                \"""uses the programs own Python API to save the current file
                as a new version.
                \"""

                # do anything that needs to be done before saving the file
                my_programs_own_python_api.save(filepath=self.version.full_path)

                # do anything that needs to be done after saving the file

    and that is it.

    The environment class by default has a property called ``version``. Holding
    the current open Version. It is None for a new scene and a
    :class:`~stalker.models.version.Version` instance in any other case.
    """

    name = "EnvironmentBase"

#    def __init__(self, name=""):
#        self._name = name
#        self._extensions = []
#        self._version = None

    def __str__(self):
        """the string representation of the environment
        """
        return self._name
    
    @property
    def version(self):
        """returns the current Version instance which is open in the
        environment
        """
        return self.get_current_version()

    @property
    def name(self):
        """returns the environment name
        """
        return self._name

    @name.setter
    def name(self, name):
        """sets the environment name
        """
        self._name = name

    def save_as(self, version):
        """The save as action of this environment. It should save the current
        scene or file to the given version.full_path
        """
        raise NotImplementedError

    def export_as(self, version):
        """Exports the contents of the open document as the given version.

        :param version: A :class:`~stalker.models.version.Version` instance
          holding the desired version.
        """
        raise NotImplementedError

    def open_(self, version, force=False):
        """the open action
        """
        raise NotImplementedError

    def post_open(self, version):
        """Runs after opening a file
        """
        raise NotImplementedError

    def import_(self, asset):
        """the import action
        """
        raise NotImplementedError

    def reference(self, asset):
        """the reference action
        """
        raise NotImplementedError
    
    def trim_repo_path(self, path):
        """Trims the repository path value from the given path

        :param path: The path that wanted to be trimmed
        :return: str
        """
        # get the repo first
        repo = self.find_repo(path)

        if not repo:
            return path

        # then try to trim the path
        if path.startswith(repo.path):
            return path[len(repo.path):]
        elif path.startswith(repo.windows_path):
            return path[len(repo.windows_path):]
        elif path.startswith(repo.linux_path):
            return path[len(repo.linux_path):]
        elif path.startswith(repo.osx_path):
            return path[len(repo.osx_path):]
        return path

    def find_repo(self, path):
        """returns the repository from the given path

        :param str path: path in a repository
        :return: stalker.models.repository.Repository
        """
        # first find the repository
        from stalker import Repository

        repos = Repository.query.all()
        found_repo = None
        for repo in repos:
            if path.startswith(repo.path) \
                or path.startswith(repo.windows_path) \
                or path.startswith(repo.linux_path) \
                or path.startswith(repo.osx_path):
                found_repo = repo
                break
        return found_repo

    def get_versions_from_path(self, path):
        """Finds Version instances from the given path value.

        Finds and returns the :class:`~stalker.models.version.Version`
        instances from the given path value.

        Returns an empty list if it can't find any matching.

        This method is different than
        :meth:`~stalker.env.EnvironmentBase.get_version_from_full_path`
        because it returns a list of
        :class:`~stalker.models.version.Version` instances which are
        residing in that path. The list is ordered by the ``id``\ s of the
        instances.

        :param path: A path which has possible
            :class:`~stalker.models.version.Version` instances.

        :return: A list of :class:`~stalker.models.version.Version` instances.
        """
        if not path:
            return []

        # convert '\\' to '/'
        path = os.path.normpath(path).replace('\\', '/')
        path_trimmed = self.trim_repo_path(path)
        logger.debug('path_trimmed: %s' % path_trimmed)

        from stalker import Version

        # try to get all versions with that info
        versions = Version.query.\
            filter(Version.full_path.startswith(path_trimmed)).all()

        return versions


    def get_version_from_full_path(self, full_path):
        """Finds the Version instance from the given full_path value.

        Finds and returns a :class:`~stalker.models.version.Version` instance
        from the given full_path value.

        Returns None if it can't find any matching.

        :param full_path: The full_path of the desired
            :class:`~stalker.models.version.Version` instance.

        :return: :class:`~stalker.models.version.Version`
        """
        # convert '\\' to '/'
        full_path = os.path.normpath(full_path).replace('\\', '/')

        # trim repo path
        full_path_trimmed = self.trim_repo_path(full_path)
        logger.debug('full_path_trimmed: %s' % full_path_trimmed)

        from stalker import Version

        # try to get a version with that info
        version = Version.query\
            .filter(Version.full_path == full_path_trimmed).first()
        return version
    
    def get_current_version(self):
        """Returns the current Version instance from the environment.

        :returns: :class:`~stalker.models.version.Version` instance or None
        """
        raise NotImplementedError

    def get_last_version(self):
        """Returns the last opened Version instance from the environment.

        * It first looks at the current open file full path and tries to match
          it with a Version instance.
        * Then searches for the recent files list.
        * Still not able to find any Version instances, will return the version
          instance with the highest id which has the current workspace path in
          its path
        * Still not able to find any Version instances returns None

        :returns: :class:`~stalker.models.version.Version` instance or None.
        """
        raise NotImplementedError

    def get_project(self):
        """returns the current project from environment
        """
        raise NotImplementedError

    def set_project(self, version):
        """Sets the project to the given Versions project.

        :param version: A :class:`~stalker.models.version.Version`.
        """
        raise NotImplementedError

#    def setOutputFileName(self):
#    def set_output_path(self):
#        """sets the output file names
#        """
#        raise NotImplementedError

#    def append_to_recent_files(self, path):
#        """appends the given path to the recent files list
#        """
#        raise NotImplementedError

    def check_referenced_versions(self):
        """Checks the referenced versions

        :returns: list of Versions
        """
        raise NotImplementedError

    def get_referenced_versions(self):
        """Returns the :class:`~stalker.models.version.Version` instances which
        are referenced in to the current scene

        :returns: list of :class:`~stalker.models.version.Version` instances.
        """
        raise NotImplementedError

#    def update_versions(self, version_tuple_list):
#        """updates the assets to the latest versions
#        """
#        raise NotImplementedError

    def get_frame_range(self):
        """Returns the frame range from the environment

        :returns: a tuple of integers containing the start and end frame
            numbers
        """
        raise NotImplementedError

    def set_frame_range(self, start_frame=1, end_frame=100,
                        adjust_frame_range=False):
        """Sets the frame range in the environment to the given start and end
        frames
        """
        raise NotImplementedError

    def get_fps(self):
        """Returns the frame rate of this current environment
        """
        raise NotImplementedError

    def set_fps(self, fps=25):
        """Sets the frame rate of the environment. The default value is 25.
        """
        raise NotImplementedError

    @property
    def extensions(self):
        """Returns the valid native extensions for this environment.

        :returns: a list of strings
        """
        return self._extensions

    @extensions.setter
    def extensions(self, extensions):
        """Sets the valid native extensions of this environment.

        :param extensions: A list of strings holding the extensions. Ex:
            ["ma", "mb"] for Maya
        """
        self._extensions = extensions

    def has_extension(self, filename):
        """Returns True if the given file names extension is in the extensions
        list false otherwise.

        accepts:
        * a full path with extension or not
        * a file name with extension or not
        * an extension with a dot on the start or not

        :param filename: A string containing the filename
        """
        if filename is None:
            return False
        return filename.split('.')[-1].lower() in self.extensions

    def load_referenced_versions(self):
        """loads all the references
        """
        raise NotImplementedError

    def replace_version(self, source_version, target_version):
        """Replaces the source_version with the target_version

        :param source_version: A
          :class:`~stalker.models.version.Version` instance holding the version
          to be replaced

        :param target_version: A
          :class:`~stalker.models.version.Version` instance holding the new
          version replacing the source one.
        """
        raise NotImplementedError

    def replace_external_paths(self, mode=0):
        """Replaces the external paths (which are not starting with the
        environment variable) with a proper path. The mode controls if the
        resultant path should be absolute or relative to the project dir.

        :param mode: Controls the resultant path is absolute or relative.

          mode 0: absolute (a path which starts with $REPO)
          mode 1: relative (to project path)

        :return:
        """
        raise NotImplementedError




# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, validates, synonym

from stalker import defaults
from stalker.models.auth import User
from stalker.models.entity import Entity

from stalker.log import logging_level
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


class Department(Entity):
    """The departments that forms the studio itself.

    The information that a Department object holds is like:

      * The members of the department
      * The lead of the department
      * and all the other things those are inherited from the AuditEntity class

    Two Department object considered the same if they have the same name, the
    the members list nor the lead info is important, a "Modeling" department
    should of course be the same with another department which has the name
    "Modeling" again.

    so creating a department object needs the following parameters:

    :param members: it can be an empty list, so one department can be created
      without any member in it. But this parameter should be a list of User
      objects.

    :param lead: this is a User object, that holds the lead information, a lead
      could be in this department but it is not forced to be also a member of
      the department. So another departments member can be a lead for another
      department. Lead attribute can not be empty or None.
    """
    __auto_name__ = False
    __tablename__ = "Departments"
    __mapper_args__ = {"polymorphic_identity": "Department"}
    department_id = Column(
        "id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True
    )

    lead_id = Column(
        "lead_id",
        Integer,
        ForeignKey("Users.id", use_alter=True, name="x")
    )

    lead = relationship(
        "User",
        uselist=False,
        primaryjoin="Department.lead_id==User.user_id",
        post_update=True,
        doc="""The lead of this department, it is a User object""",
    )

    members = relationship(
        "User",
        secondary='User_Departments',
        back_populates="departments",
        doc="""List of users representing the members of this department.""",
    )
    
    users = synonym('members')

    def __init__(self, members=None, lead=None, **kwargs):
        super(Department, self).__init__(**kwargs)

        if members is None:
            members = []

        self.members = members
        self.lead = lead
    
    def __eq__(self, other):
        """the equality operator
        """
        return super(Department, self).__eq__(other) and\
               isinstance(other, Department) 
    
    @validates("members")
    def _validate_members(self, key, member):
        """validates the given member attribute
        """

        if not isinstance(member, User):
            raise TypeError("Every element in the %s.members list should be "
                            "an instance of stalker.models.auth.User class "
                            "not %s" %
                            (self.__class__.__name__,
                             member.__class__.__name__))
        return member

    @validates("lead")
    def _validate_lead(self, key, lead):
        """validates the given lead attribute
        """
        from stalker.models.auth import User
        if lead is not None:
            # the lead should be an instance of User class
            if not isinstance(lead, User):
                raise TypeError("%s.lead should be an instance of "
                                "stalker.models.auth.User not %s" %
                                (self.__class__.__name__,
                                 lead.__class__.__name__))
        return lead

    @property
    def to_tjp(self):
        """outputs a TaskJuggler compatible string
        """
        from jinja2 import Template
        temp = Template(defaults.tjp_department_template)
        return temp.render({'department': self})



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, validates

from stalker.models.mixins import ReferenceMixin, CodeMixin
from stalker.models.task import Task

from stalker.log import logging_level
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


class Sequence(Task, CodeMixin):
    """Stores data about Sequences.

    Sequences are a way of grouping the Shots according to their temporal
    position to each other.

    **Initialization**

    .. warning::

       .. deprecated:: 0.2.0

       Sequences do not have a lead anymore. Use the
       :class:`~stalker.models.task.Task.responsible` attribute of the super
       (:class:`~stalker.models.task.Task`).
    """
    __auto_name__ = False
    __tablename__ = "Sequences"
    __mapper_args__ = {"polymorphic_identity": "Sequence"}
    sequence_id = Column("id", Integer, ForeignKey("Tasks.id"),
                         primary_key=True)

    shots = relationship(
        "Shot",
        secondary='Shot_Sequences',
        back_populates="sequences",
        doc="""The :class:`~stalker.models.shot.Shot`\ s assigned to this Sequence.

        It is a list of :class:`~stalker.models.shot.Shot` instances.
        """
    )

    def __init__(self, **kwargs):
        super(Sequence, self).__init__(**kwargs)

        # call the mixin __init__ methods
        ReferenceMixin.__init__(self, **kwargs)
        CodeMixin.__init__(self, **kwargs)
        self.shots = []

    @validates("shots")
    def _validate_shots(self, key, shot):
        """validates the given shot value
        """
        from stalker.models.shot import Shot

        if not isinstance(shot, Shot):
            raise TypeError('%s.shots should be all '
                            'stalker.models.shot.Shot instances, not %s')
        return shot

    def __eq__(self, other):
        """the equality operator
        """
        return isinstance(other, Sequence) and \
            super(Sequence, self).__eq__(other)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from sqlalchemy import Column, Integer, ForeignKey
from stalker.models.entity import Entity
from stalker.models.mixins import StatusMixin

from stalker.log import logging_level
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


class Message(Entity, StatusMixin):
    """The base of the messaging system in Stalker

    Messages are one of the ways to collaborate in Stalker. The model of the
    messages is taken from the e-mail system. So it is pretty similar to an
    e-mail message.

    :param from: the :class:`~stalker.models.user.User` object sending the
      message

    :param to: the list of :class:`~stalker.models.user.User`\ s to
      receive this message

    :param subject: the subject of the message

    :param body: the body of the message

    :param in_reply_to: the :class:`~stalker.models.message.Message`
      object which this message is a reply to.

    :param replies: the list of :class:`~stalker.models.message.Message`
      objects which are the direct replies of this message

    :param attachments: a list of
      :class:`~stalker.models.entity.SimpleEntity` objects attached to
      this message (so anything can be attached to a message)

    """
    __auto_name__ = True
    __tablename__ = "Messages"
    __mapper_args__ = {"polymorphic_identity": "Message"}
    message_id = Column("id", Integer, ForeignKey("Entities.id"),
                        primary_key=True)

    def __init__(self, **kwargs):
        super(Message, self).__init__(**kwargs)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import datetime
import logging

from sqlalchemy import (Table, Column, String, Integer, ForeignKey, Interval,
                        DateTime, PickleType)
from sqlalchemy.exc import UnboundExecutionError
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import synonym, relationship, validates

from stalker import defaults
from stalker.db import Base
from stalker.db.session import DBSession
from stalker.log import logging_level
from stalker.models import make_plural

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


def create_secondary_table(
        primary_cls_name,
        secondary_cls_name,
        primary_cls_table_name,
        secondary_cls_table_name,
        secondary_table_name=None
):
    """creates any secondary table
    """

    plural_secondary_cls_name = make_plural(secondary_cls_name)

    # use the given class_name and the class_table
    if not secondary_table_name:
        secondary_table_name = \
            primary_cls_name + "_" + plural_secondary_cls_name

    # check if the table is already defined
    if secondary_table_name not in Base.metadata:
        secondary_table = Table(
            secondary_table_name, Base.metadata,
            Column(
                primary_cls_name.lower() + "_id",
                Integer,
                ForeignKey(primary_cls_table_name + ".id"),
                primary_key=True,
            ),

            Column(
                secondary_cls_name.lower() + "_id",
                Integer,
                ForeignKey(secondary_cls_table_name + ".id"),
                primary_key=True,
            )
        )
    else:
        secondary_table = Base.metadata.tables[secondary_table_name]

    return secondary_table


class TargetEntityTypeMixin(object):
    """Adds target_entity_type attribute to mixed in class.

    :param target_entity_type: The target entity type which this class is 
      designed for. Should be a class or a class name.

      For example::

        from stalker import SimpleEntity, TargetEntityTypeMixin, Project

        class A(SimpleEntity, TargetEntityTypeMixin):
            __tablename__ = "As"
            __mapper_args__ = {"polymorphic_identity": "A"}

            def __init__(self, **kwargs):
                super(A, self).__init__(**kwargs)
                TargetEntityTypeMixin.__init__(self, **kwargs)

        a_obj = A(target_entity_type=Project)

      The ``a_obj`` will only be accepted by
      :class:`~stalker.models.project.Project` instances. You can not assign it
      to any other class which accepts a :class:`~stalker.models.type.Type`
      instance.

    To control the mixed-in class behaviour add these class variables to the 
    mixed in class:

      __nullable_target__ : controls if the target_entity_type can be 
                            nullable or not. Default is False.

      __unique_target__ : controls if the target_entity_type should be 
                          unique, so there is only one object for one type.
                          Default is False.
    """

    __nullable_target__ = False
    __unique_target__ = False

    @declared_attr
    def _target_entity_type(cls):
        return Column(
            "target_entity_type",
            String(128),
            nullable=cls.__nullable_target__,
            unique=cls.__unique_target__
        )

    def __init__(self, target_entity_type=None, **kwargs):
        self._target_entity_type = \
            self._validate_target_entity_type(target_entity_type)

    def _validate_target_entity_type(self, target_entity_type_in):
        """validates the given target_entity_type value
        """
        # it can not be None
        if target_entity_type_in is None:
            raise TypeError("%s.target_entity_type can not be None" %
                            self.__class__.__name__)

        if str(target_entity_type_in) == "":
            raise ValueError("%s.target_entity_type can not be empty" %
                             self.__class__.__name__)

        # check if it is a class
        if isinstance(target_entity_type_in, type):
            target_entity_type_in = target_entity_type_in.__name__

        return str(target_entity_type_in)

    def _target_entity_type_getter(self):
        return self._target_entity_type

    @declared_attr
    def target_entity_type(cls):
        return synonym(
            "_target_entity_type",
            descriptor=property(
                fget=cls._target_entity_type_getter,
                doc="""The entity type which this object is valid for.

                Usually it is set to the TargetClass directly.
                """
            )
        )


class StatusMixin(object):
    """Makes the mixed in object statusable.

    This mixin adds status and status_list attributes to the mixed in class.
    Any object that needs a status and a corresponding status list can include
    this mixin.

    When mixed with a class which don't have an __init__ method, the mixin
    supplies one, and in this case the parameters below must be defined.

    :param status_list: this attribute holds a status list object, which shows
      the possible statuses that this entity could be in. This attribute can
      not be empty or None. Giving a StatusList object, the
      StatusList.target_entity_type should match the current class.

      .. versionadded:: 0.1.2.a4

        The status_list argument now can be skipped or can be None if there
        is an active database connection (stalker.models.DBSession is not
        None) and there is a suitable
        :class:`~stalker.models.status.StatusList` instance in the database
        whom :attr:`~stalker.models.status.StatusList.target_entity_type`
        attribute is set to the current mixed-in class name.

    :param status: It is a :class:`~stalker.models.status.Status` instance
      which shows the current status of the statusable object. Integer values
      are also accepted, which shows the index of the desired status in the
      ``status_list`` attribute of the current statusable object. If a
      :class:`~stalker.models.status.Status` instance is supplied, it should
      also be present in the ``status_list`` attribute. If set to None then the
      first :class:`~stalker.models.status.Status` instance in the
      ``status_list`` will be used.

      .. versionadded:: 0.2.0

        Status attribute as Status instance:

        It is now possible to set the status of the instance by a
        :class:`~stalker.models.status.Status` instance directly. And the
        :attr:`~stalker.models.mixins.StatusMixin.status` will return a proper
        :class:`~stalker.models.status.Status` instance.
    """

    def __init__(self, status=None, status_list=None, **kwargs):
        self.status_list = status_list
        self.status = status
        # logger.debug('%s.status: %s' % (self.__class__.__name__, status))

    @declared_attr
    def status_id(cls):
        return Column(
            'status_id',
            Integer,
            ForeignKey('Statuses.id'),
            nullable=False
            # This is set to nullable=True but it is impossible to set the
            # status to None by using this Declarative approach.
            # 
            # This is done in that way cause SQLAlchemy was flushing the data
            # (AutoFlush) preliminarily while checking if the given Status was
            # in the related StatusList, and it was complaining about the
            # status can not be null
        )

    @declared_attr
    def status(cls):
        return relationship(
            'Status',
            primaryjoin= \
                "%s.status_id==Status.status_id" % cls.__name__,
            doc="""The current status of the object.

            It is a :class:`~stalker.models.status.Status` instance which
            is one of the Statuses stored in the ``status_list`` attribute
            of this object.
            """
        )

    @declared_attr
    def status_list_id(cls):
        return Column(
            'status_list_id',
            Integer,
            ForeignKey('StatusLists.id'), #, use_alter=True, name="x"),
            nullable=False
        )

    @declared_attr
    def status_list(cls):
        return relationship(
            "StatusList",
            primaryjoin= \
                "%s.status_list_id==StatusList.status_list_id" %
                cls.__name__,
        )

    @validates("status_list")
    def _validate_status_list(self, key, status_list):
        """validates the given status_list_in value
        """
        from stalker.models.status import StatusList

        if status_list is None:
            # check if there is a db setup and try to get the appropriate 
            # StatusList from the database

            # disable autoflush to prevent premature class initialization
            with DBSession.no_autoflush:
                try:
                    # try to get a StatusList with the target_entity_type is 
                    # matching the class name
                    status_list = DBSession.query(StatusList) \
                        .filter_by(target_entity_type=self.__class__.__name__) \
                        .first()
                except UnboundExecutionError:
                    # it is not mapped just skip it
                    pass

        # if it is still None
        if status_list is None:
            # there is no db so raise an error because there is no way 
            # to get an appropriate StatusList
            raise TypeError(
                "%s instances can not be initialized without a "
                "stalker.models.status.StatusList instance, please pass a "
                "suitable StatusList (StatusList.target_entity_type=%s) "
                "with the 'status_list' argument" %
                (self.__class__.__name__, self.__class__.__name__)
            )
        else:
            # it is not an instance of status_list
            if not isinstance(status_list, StatusList):
                raise TypeError(
                    "%s.status_list should be an instance of "
                    "stalker.models.status.StatusList not %s" %
                    (self.__class__.__name__,
                     status_list.__class__.__name__)
                )

            # check if the entity_type matches to the StatusList.target_entity_type
            if self.__class__.__name__ != status_list.target_entity_type:
                raise TypeError(
                    "the given StatusLists' target_entity_type is %s, "
                    "whereas the entity_type of this object is %s" % \
                    (status_list.target_entity_type,
                     self.__class__.__name__))

        return status_list

    @validates('status')
    def _validate_status(self, key, status):
        """validates the given status value
        """
        from stalker.models.status import Status, StatusList

        if not isinstance(self.status_list, StatusList):
            raise TypeError("please set the %s.status_list attribute first" %
                            self.__class__.__name__)

        # it is set to None
        if status is None:
            with DBSession.no_autoflush:
                status = self.status_list.statuses[0]

        # it is not an instance of status or int
        if not isinstance(status, (Status, int)):
            raise TypeError("%s.status must be an instance of "
                            "stalker.models.status.Status or an integer "
                            "showing the index of the Status object in the "
                            "%s.status_list, not %s" %
                            (self.__class__.__name__,
                             self.__class__.__name__,
                             status.__class__.__name__))

        if isinstance(status, int):
            # if it is not in the correct range:
            if status < 0:
                raise ValueError("%s.status must be a non-negative integer" %
                                 self.__class__.__name__)

            if status >= len(self.status_list.statuses):
                raise ValueError("%s.status can not be bigger than the length "
                                 "of the status_list" %
                                 self.__class__.__name__)
                # get the tatus instance out of the status_list instance
            status = self.status_list[status]

        # check if the given status is in the status_list
        # logger.debug('self.status_list: %s' % self.status_list)
        # logger.debug('given status: %s' % status)

        if status not in self.status_list:
            raise ValueError("The given Status instance for %s.status is not "
                             "in the %s.status_list, please supply a status "
                             "from that list." %
                             (self.__class__.__name__, self.__class__.__name__)
            )

        return status


class ScheduleMixin(object):
    """Adds schedule info to the mixed in class.

    Adds schedule information like ``start``, ``end`` and ``duration``. These
    attributes will be used in TaskJuggler. Because ``effort`` is only
    meaningful if there are some ``resources`` this attribute has been left
    special for :class:`~stalker.models.task.Task` class. The ``length`` has
    not been implemented because of its rare use.

    The preceding order for the attributes is as follows::

      start > end > duration

    So if all of the parameters are given only the ``start`` and the ``end``
    will be used and the ``duration`` will be calculated accordingly. In any
    other conditions the missing parameter will be calculated from the
    following table:

    +-------+-----+----------+----------------------------------------+
    | start | end | duration | DEFAULTS                               |
    +=======+=====+==========+========================================+
    |       |     |          | start = datetime.datetime.now()        |
    |       |     |          |                                        |
    |       |     |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |     |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |  X  |          | duration = end - start                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |     |    X     | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |  X  |    X     | duration = end - start                 |
    +-------+-----+----------+----------------------------------------+
    |       |  X  |    X     | start = end - duration                 |
    +-------+-----+----------+----------------------------------------+
    |       |  X  |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | start = end - duration                 |
    +-------+-----+----------+----------------------------------------+
    |       |     |    X     | start = datetime.datetime.now()        |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+

    Only the ``start``, ``end`` will be stored. The ``duration`` attribute is
    the direct difference of the the ``start`` and ``end`` attributes, so there
    is no need to store it. But if will be used in calculation of the start and
    end values.

    The start and end attributes have a ``computed`` companion. Which are the
    return values from TaskJuggler. so for start there is the
    ``computed_start`` and for end there is the ``computed_end`` attributes.
    These values are going to be used in Gantt Charts.

    The date attributes can be managed with timezones. Follow the Python idioms
    shown in the `documentation of datetime`_

    .. _documentation of datetime: http://docs.python.org/library/datetime.html

    :param start: the start date of the entity, should be a datetime.datetime
      instance, the start is the pin point for the date calculation. In
      any condition if the start is available then the value will be
      preserved. If start passes the end the end is also changed
      to a date to keep the timedelta between dates. The default value is
      datetime.datetime.now()

    :type start: :class:`datetime.datetime`

    :param end: the end of the entity, should be a datetime.datetime instance,
      when the start is changed to a date passing the end, then the end is also
      changed to a later date so the timedelta between the dates is kept.

    :type end: :class:`datetime.datetime` or :class:`datetime.timedelta`

    :param duration: The duration of the entity. It is a
      :class:`datetime.timedelta` instance. The default value is read from
      the :class:`~stalker.config.Config` class. See the table above for the
      initialization rules.

    :type duration: :class:`datetime.timedelta`

    :param timing_resolution: The timing_resolution of the datetime.datetime
      object in datetime.timedelta. Uses ``timing_resolution`` settings in the
      :class:`stalker.config.Config` class which defaults to 1 hour. Setting
      the timing_resolution to less then 5 minutes is not suggested because it
      is a limit for TaskJuggler.

    :type timing_resolution: datetime.timedelta
    """

    #    # add this lines for Sphinx
    #    __tablename__ = "ScheduleMixins"

    def __init__(self,
                 start=None,
                 end=None,
                 duration=None,
                 timing_resolution=None,
                 **kwargs):
        self.timing_resolution = timing_resolution
        self._validate_dates(start, end, duration)

    @declared_attr
    def _end(cls):
        return Column("end", DateTime)

    def _end_getter(self):
        """The date that the entity should be delivered.

        The end can be set to a datetime.timedelta and in this case it will be
        calculated as an offset from the start and converted to
        datetime.datetime again. Setting the start to a date passing the end
        will also set the end, so the timedelta between them is preserved,
        default value is 10 days
        """
        return self._end

    def _end_setter(self, end_in):
        self._validate_dates(self.start, end_in, self.duration)

    @declared_attr
    def end(cls):
        return synonym(
            "_end",
            descriptor=property(
                cls._end_getter,
                cls._end_setter
            )
        )

    @declared_attr
    def _start(cls):
        return Column("start", DateTime)

    def _start_getter(self):
        """The date that this entity should start.

        Also effects the
        :attr:`~stalker.models.mixins.ScheduleMixin.end` attribute value in
        certain conditions, if the
        :attr:`~stalker.models.mixins.ScheduleMixin.start` is set to a time
        passing the :attr:`~stalker.models.mixins.ScheduleMixin.end` it will
        also offset the :attr:`~stalker.models.mixins.ScheduleMixin.end` to
        keep the :attr:`~stalker.models.mixins.ScheduleMixin.duration` value
        fixed. :attr:`~stalker.models.mixins.ScheduleMixin.start` should be an
        instance of class:`datetime.datetime` and the default value is
        :func:`datetime.datetime.now()`
        """
        return self._start

    def _start_setter(self, start_in):
        self._validate_dates(start_in, self.end, self.duration)

    @declared_attr
    def start(cls):
        return synonym(
            "_start",
            descriptor=property(
                cls._start_getter,
                cls._start_setter,
            )
        )

    @declared_attr
    def _duration(cls):
        return Column('duration', Interval)

    def _duration_getter(self):
        return self._duration

    def _duration_setter(self, duration_in):
        if duration_in is not None:
            if isinstance(duration_in, datetime.timedelta):
                # set the end to None
                # to make it recalculated
                self._validate_dates(self.start, None, duration_in)
            else:
                self._validate_dates(self.start, self.end, duration_in)
        else:
            self._validate_dates(self.start, self.end, duration_in)

    @declared_attr
    def duration(self):
        return synonym(
            '_duration',
            descriptor=property(
                self._duration_getter,
                self._duration_setter,
                doc="""Duration of the entity.

It is a datetime.timedelta instance. Showing the difference of the
:attr:`.start` and the :attr:`.end`. If edited it changes the :attr:`.end`
attribute value."""
            )

        )

    def _validate_dates(self, start, end, duration):
        """updates the date values
        """
        # logger.debug('start    : %s' % start)
        # logger.debug('end      : %s' % end)
        # logger.debug('duration : %s' % duration)

        if not isinstance(start, datetime.datetime):
            start = None

        if not isinstance(end, datetime.datetime):
            end = None

        if not isinstance(duration, datetime.timedelta):
            duration = None

        # check start
        if start is None:
            # try to calculate the start from end and duration
            if end is None:
                # set the defaults
                start = datetime.datetime.now()

                if duration is None:
                    # set the defaults
                    duration = defaults.task_duration

                end = start + duration
            else:
                if duration is None:
                    duration = defaults.task_duration

                start = end - duration

        # check end
        if end is None:
            if duration is None:
                duration = defaults.task_duration

            end = start + duration

        if end < start:
            # check duration
            if duration is None or duration < datetime.timedelta(1):
                duration = datetime.timedelta(1)

            end = start + duration

        # round the dates to the timing_resolution
        self._start = self.round_time(start)
        self._end = self.round_time(end)
        self._duration = self._end - self._start

    @declared_attr
    def _timing_resolution(cls):
        return Column("timing_resolution", Interval)

    def _timing_resolution_getter(self):
        """returns the timing_resolution
        """
        return self._timing_resolution

    def _timing_resolution_setter(self, res_in):
        """sets the timing_resolution
        """
        self._timing_resolution = self._validate_timing_resolution(res_in)
        logger.debug('self._timing_resolution: %s' % self._timing_resolution)
        # update date values
        if self.start and self.end and self.duration:
            self._validate_dates(
                self.round_time(self.start),
                self.round_time(self.end),
                None
            )

    @declared_attr
    def timing_resolution(cls):
        return synonym(
            '_timing_resolution',
            descriptor=property(
                cls._timing_resolution_getter,
                cls._timing_resolution_setter,
                doc="""The timing_resolution of this object.

                Can be set to any value that is representable with
                datetime.timedelta. The default value is 1 hour. Whenever it is
                changed the start, end and duration values will be updated.
                """
            )
        )

    def _validate_timing_resolution(self, timing_resolution):
        """validates the given timing_resolution value
        """
        if timing_resolution is None:
            timing_resolution = defaults.timing_resolution

        if not isinstance(timing_resolution, datetime.timedelta):
            raise TypeError('%s.timing_resolution should be an instance of '
                            'datetime.timedelta not, %s' %
                            (self.__class__.__name__,
                             timing_resolution.__class__.__name__))

        return timing_resolution

    @declared_attr
    def computed_start(cls):
        return Column('computed_start', DateTime)

    @declared_attr
    def computed_end(cls):
        return Column('computed_end', DateTime)


    @property
    def computed_duration(self):
        """returns the computed_duration as the difference of computed_start
        and computed_end if there are computed_start and computed_end otherwise
        returns None
        """
        return self.computed_end - self.computed_start \
            if self.computed_end and self.computed_start else None

    def round_time(self, dt):
        """Round a datetime object to any time laps in seconds.

        Uses class property timing_resolution as the closest number of seconds
        to round to, defaults 1 hour.

        :param dt: datetime.datetime object, defaults now.

        Based on Thierry Husson's answer in `Stackoverflow`_

        _`Stackoverflow` : http://stackoverflow.com/a/10854034/1431079
        """
        # to be compatible with python 2.6 use the following instead of
        # total_seconds()
        trs = self.timing_resolution.days * 86400 + \
             self.timing_resolution.seconds

        # convert to seconds
        # FIX: using strftime(%s) is dangerous, it uses system time zone
        epoch = datetime.datetime(1970, 1, 1)
        diff = dt - epoch
        diff_in_seconds = diff.days * 86400 + diff.seconds
        return epoch + datetime.timedelta(
            seconds=(diff_in_seconds + trs * 0.5) // trs * trs
        )

    @property
    def total_seconds(self):
        """returns the duration as seconds
        """
        return self.duration.days * 86400 + self.duration.seconds

    @property
    def computed_total_seconds(self):
        """returns the duration as seconds
        """
        return self.computed_duration.days * 86400 + \
               self.computed_duration.seconds


class ProjectMixin(object):
    """Gives the ability to connect to a :class:`~stalker.models.project.Project` to the mixed in object.

    :param project: A :class:`~stalker.models.project.Project` instance holding
      the project which this object is related to. It can not be None, or
      anything other than a :class:`~stalker.models.project.Project` instance.

    :type project: :class:`~stalker.models.project.Project`
    """

    #    # add this lines for Sphinx
    #    __tablename__ = "ProjectMixins"

    @declared_attr
    def project_id(cls):
        return Column(
            "project_id",
            Integer,
            ForeignKey("Projects.id", use_alter=True, name="project_x_id"),
            #ForeignKey("Projects.id"),
            # cannot use nullable cause a Project object needs
            # insert itself as the project and it needs post_update
            # thus nullable should be True
            #nullable=False,
        )

    @declared_attr
    def project(cls):
        backref = cls.__tablename__.lower()
        doc = """The :class:`~stalker.models.project.Project` instance that
        this object belongs to.
        """

        return relationship(
            "Project",
            primaryjoin= \
                cls.__tablename__ + ".c.project_id==Projects.c.id",
            post_update=True, # for project itself
            uselist=False,
            backref=backref,
            doc=doc
        )

    def __init__(self,
                 project=None,
                 **kwargs):
        self.project = project

    @validates("project")
    def _validate_project(self, key, project):
        """validates the given project value
        """

        from stalker.models.project import Project

        if project is None:
            raise TypeError("%s.project can not be None it must be an "
                            "instance of stalker.models.project.Project" %
                            self.__class__.__name__)

        if not isinstance(project, Project):
            raise TypeError("%s.project should be an instance of "
                            "stalker.models.project.Project instance not %s" %
                            (self.__class__.__name__,
                             project.__class__.__name__))
        return project


class ReferenceMixin(object):
    """Adds reference capabilities to the mixed in class.

    References are :class:`stalker.models.link.Link` instances or anything
    derived from it, which adds information to the attached objects. The aim of
    the References are generally to give more info to direct the evolution of
    the object.

    :param references: A list of :class:`~stalker.models.link.Link` instances.

    :type references: list of :class:`~stalker.models.link.Link` instances.
    """
    # add this lines for Sphinx
    #    __tablename__ = "ReferenceMixins"

    def __init__(self,
                 references=None,
                 **kwargs):
        if references is None:
            references = []

        self.references = references

    @declared_attr
    def references(cls):
        # get secondary table
        secondary_table = create_secondary_table(
            cls.__name__,
            'Link',
            cls.__tablename__,
            'Links',
            cls.__name__ + "_References"
        )
        # return the relationship
        return relationship(
            "Link",
            secondary=secondary_table,
            doc="""A list of :class:`~stalker.models.link.Link` instances given
            as a reference for this entity.
            """
        )

    @validates("references")
    def _validate_references(self, key, reference):
        """validates the given reference
        """

        from stalker.models.entity import SimpleEntity

        # all the elements should be instance of stalker.models.entity.Entity
        if not isinstance(reference, SimpleEntity):
            raise TypeError("%s.references should be all instances of "
                            "stalker.models.entity.SimpleEntity not %s"
                            % (self.__class__.__name__,
                               reference.__class__.__name__))
        return reference


class ACLMixin(object):
    """A Mixin for adding ACLs to mixed in class.

    Access control lists or ACLs are used to determine if the given resource
    has the permission to access the given data. It is based on Pyramids
    Authorization system but organized to fit in Stalker style.

    The ACLMixin adds an attribute called ``permissions`` and a
    property called ``__acl__`` to be able to pass the permission data to
    Pyramid framework.
    """

    @declared_attr
    def permissions(cls):
        # get the secondary table
        secondary_table = create_secondary_table(
            cls.__name__, 'Permission', cls.__tablename__, 'Permissions'
        )
        return relationship('Permission', secondary=secondary_table)

    @validates('permissions')
    def _validate_permissions(self, key, permission):
        """validates the given permission value
        """
        from stalker.models.auth import Permission

        if not isinstance(permission, Permission):
            raise TypeError("%s.permissions should be all instances of "
                            "stalker.models.auth.Permission not %s" %
                            (self.__class__.__name__,
                             permission.__class__.__name__))

        return permission

    @property
    def __acl__(self):
        """Returns Pyramid friendly ACL list composed by the:

          * Permission.access (Ex: 'Allow' or 'Deny')
          * The Mixed in class name and the object name (Ex: 'User:eoyilmaz')
          * The Action and the target class name (Ex: 'Create_Asset')

        Thus a list of tuple is returned as follows::

          __acl__ = [
              ('Allow', 'User:eoyilmaz', 'Create_Asset'),
          ]

        For the last example user eoyilmaz can grant access to views requiring
        'Add_Project' permission.
        """
        return [(perm.access,
                 self.__class__.__name__ + ':' + self.name,
                 perm.action + '_' + perm.class_name)
                for perm in self.permissions]


class CodeMixin(object):
    """Adds code info to the mixed in class.

    .. versionadded:: 0.2.0

      The code attribute of the SimpleEntity is now introduced as a separate
      mixin. To let it be used by the classes it is really needed. 

    The CodeMixin just adds a new field called ``code``. It is a very simple
    attribute and is used for simplifying long names (like Project.name etc.).

    Contrary to previous implementations the code attribute is not formatted in
    anyway, so care needs to be taken if the code attribute is going to be used
    in filesystem as file and directory names.

    :param str code: The code attribute is a string, can not be empty or can
      not be None.
    """

    def __init__(
            self,
            code=None,
            **kwargs):
        logger.debug('code: %s' % code)
        self.code = code

    @declared_attr
    def code(cls):
        return Column(
            'code',
            String(256),
            nullable=False,
            doc="""The code name of this object.

                It accepts strings. Can not be None."""
        )

    @validates('code')
    def _validate_code(self, key, code):
        """validates the given code attribute
        """
        logger.debug('validating code value of: %s' % code)
        if code is None:
            raise TypeError("%s.code cannot be None" % self.__class__.__name__)

        if not isinstance(code, (str, unicode)):
            raise TypeError('%s.code should be an instance of string or '
                            'unicode not %s' %
                            (self.__class__.__name__,
                             code.__class__.__name__)
            )

        if code == '':
            raise ValueError('%s.code can not be an empty string')

        return code


class WorkingHoursMixin(object):
    """Sets working hours for the mixed in class.

    Generally is meaningful for users, departments and studio.

    :param working_hours: A :class:`~stalker.models.project.WorkingHours`
      instance showing the working hours settings for that project. This data
      is stored as a PickleType in the database.
    """

    def __init__(self,
                 working_hours=None,
                 **kwargs):
        self.working_hours = working_hours

    @declared_attr
    def working_hours(cls):
        return Column(PickleType)

    @validates('working_hours')
    def _validate_working_hours(self, key, wh):
        """validates the given working hours value
        """
        if wh is None:
            # use the default one
            from stalker import WorkingHours

            wh = WorkingHours()
        return wh



#-*- coding: utf-8 -*-

import unittest2
from stalker import Entity, Note, Tag, User

class EntityTester(unittest2.TestCase):
    """tests the Entity class
    """
    
    def setUp(self):
        """setting up some proper values
        """
        # create a user
        self.test_user = User(
            name="Test User",
            login="testuser",
            email="test@user.com",
            password="test"
        )

        # create some test Tag objects, not necessarily needed but create them
        self.test_tag1 = Tag(name="Test Tag 1")
        self.test_tag2 = Tag(name="Test Tag 1") # make it equal to tag1
        self.test_tag3 = Tag(name="Test Tag 3")

        self.tags = [self.test_tag1, self.test_tag2]

        # create a couple of test Note objects
        self.test_note1 = Note(name="test note1", content="test note1")
        self.test_note2 = Note(name="test note2", content="test note2")
        self.test_note3 = Note(name="test note3", content="test note3")

        self.notes = [self.test_note1, self.test_note2]

        self.kwargs = {
            "name": "Test Entity",
            "description": "This is a test entity, and this is a proper \
            description for it",
            "created_by": self.test_user,
            "updated_by": self.test_user,
            "tags": self.tags,
            "notes": self.notes,
            }

        # create a proper SimpleEntity to use it later in the tests
        self.test_entity = Entity(**self.kwargs)
    

    def test___auto_name__class_attribute_is_set_to_True(self):
        """testing if the __auto_name__ class attribute is set to False for
        Entity class
        """
        self.assertTrue(Entity.__auto_name__)
     
    def test_notes_argument_being_omitted(self):
        """testing if no error raised when omitted the notes argument
        """
        self.kwargs.pop("notes")
        new_entity = Entity(**self.kwargs)

    def test_notes_argument_is_set_to_None(self):
        """testing if the notes attribute will be set to an empty list when the
        notes argument is set to None
        """
        self.kwargs["notes"] = None
        new_entity = Entity(**self.kwargs)
        self.assertEqual(new_entity.notes, [])

    def test_notes_attribute_is_set_to_None(self):
        """testing if a TypeError will be raised when the notes attribute is
        set to None
        """
        self.assertRaises(TypeError, setattr, self.test_entity, "notes", None)

    def test_notes_argument_set_to_something_other_than_a_list(self):
        """testing if a TypeError will be raised when setting the notes
        argument something other than a list
        """
        test_values = [1, 1.2, "a string note"]
        for test_value in test_values:
            self.kwargs["notes"] = test_value
            self.assertRaises(TypeError, Entity, **self.kwargs)

    def test_notes_attribute_set_to_something_other_than_a_list(self):
        """testing if a TypeError will be raised when setting the notes
        argument something other than a list
        """
        test_values = [1, 1.2, "a string note"]
        for test_value in test_values:
            self.assertRaises(
                TypeError,
                setattr,
                self.test_entity,
                "notes",
                test_value
            )

    def test_notes_argument_set_to_a_list_of_other_objects(self):
        """testing if a TypeError will be raised when trying to set the notes
        argument to a list of other kind of objects than Note objects
        """
        self.kwargs["notes"] = [1, 12.2, "this is a string",
            ["a list"], {"a": "note"}]
        self.assertRaises(TypeError, Entity, **self.kwargs)

    def test_notes_attribute_set_to_a_list_of_other_objects(self):
        """testing if a TypeError will be raised when trying to set the notes
        attribute to a list of other kind of objects than Note objects
        """
        test_value = [1, 12.2, "this is a string", ["a list"], {"a": "note"}]
        self.assertRaises(
            TypeError,
            setattr,
            self.test_entity,
            "notes",
            test_value
        )

    def test_notes_attribute_works_properly(self):
        """testing if the notes attribute works properly
        """
        test_value = [self.test_note3]
        self.test_entity.notes = test_value
        self.assertEqual(self.test_entity.notes, test_value)

    def test_notes_attribute_element_is_set_to_something_other_than_a_note_object(self):
        """testing if a TypeError will be raised when trying to assign an
        element to the notes list which is not an instance of Note
        """
        self.assertRaises(
            TypeError,
            self.test_entity.notes.__setitem__,
            0,
            0
        )

    def test_tags_argument_being_omitted(self):
        """testing if nothing is raised when creating an entity without setting
        a tags argument
        """
        self.kwargs.pop("tags")
        # this should work without errors
        aNewEntity = Entity(**self.kwargs)

    def test_tags_argument_being_initialized_as_an_empty_list(self):
        """testing if nothing happens when tags argument an empty list
        """
        # this should work without errors
        self.kwargs.pop("tags")
        aNewEntity = Entity(**self.kwargs)
        expected_result = []
        self.assertEqual(aNewEntity.tags, expected_result)

    def test_tags_argument_set_to_something_other_than_a_list(self):
        """testing if a TypeError is going to be raised when initializing the
        tags with something other than a list
        """
        test_values = ["a tag", 1243, 12.12, {"a": "tag"}]
        for test_value in test_values:
            self.kwargs["tags"] = test_value
            self.assertRaises(TypeError, Entity, **self.kwargs)

    def test_tags_attribute_works_properly(self):
        """testing if tags attribute works properly
        """
        test_value = [self.test_tag1]
        self.test_entity.tags = test_value
        self.assertEqual(self.test_entity.tags, test_value)

    def test_tags_attribute_element_is_set_to_something_other_than_a_tag_object(self):
        """testing if a TypeError will be raised when trying to assign an
        element to the tags list which is not an instance of Tag
        """
        self.assertRaises(
            TypeError,
            self.test_entity.tags.__setitem__,
            0,
            0
        )

    def test_tags_attribute_set_to_None(self):
        """testing if a TypeError will be raised when the tags attribute is set
        to None
        """
        self.assertRaises(TypeError, setattr, self.test_entity, "tags", None)

    def test_equality(self):
        """testing equality of two entities
        """
        # create two entities with same parameters and check for equality
        entity1 = Entity(**self.kwargs)
        entity2 = Entity(**self.kwargs)

        self.kwargs["name"] = "another entity"
        self.kwargs["tags"] = [self.test_tag3]
        self.kwargs["notes"] = []
        entity3 = Entity(**self.kwargs)

        self.assertTrue(entity1 == entity2)
        self.assertFalse(entity1 == entity3)

    def test_inequality(self):
        """testing inequality of two entities
        """
        # change the tags and test it again, expect False
        entity1 = Entity(**self.kwargs)
        entity2 = Entity(**self.kwargs)

        self.kwargs["name"] = "another entity"
        self.kwargs["tags"] = [self.test_tag3]
        self.kwargs["notes"] = []
        entity3 = Entity(**self.kwargs)

        self.assertFalse(entity1 != entity2)
        self.assertTrue(entity1 != entity3)





# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest2
from stalker import Entity, Status
from stalker import config
defaults = config.Config()

class StatusTest(unittest2.TestCase):
    """tests the stalker.models.status.Status class
    """
    
    def setUp(self):
        """setup the test
        """
        self.kwargs = {
            'name': 'Complete',
            'description': 'use this when the object is complete',
            'code': 'CMPLT',
            'fg_color': 0x000000,
            'bg_color': 0x00ff00,
        }
        
        # create an entity object with same kwargs for __eq__ and __ne__ tests
        # (it should return False for __eq__ and True for __ne__ for same
        # kwargs)
        self.entity1 = Entity(**self.kwargs)
    
    def test___auto_name__class_attribute_is_set_to_False(self):
        """testing if the __auto_name__ class attribute is set to False for
        Status class
        """
        self.assertFalse(Status.__auto_name__)
     
    def test_equality(self):
        """testing equality of two statuses
        """
        status1 = Status(**self.kwargs)
        status2 = Status(**self.kwargs)

        self.kwargs["name"] = "Work In Progress"
        self.kwargs["description"] = "use this when the object is still in \
        progress"
        self.kwargs["code"] = "WIP"

        status3 = Status(**self.kwargs)

        self.assertTrue(status1 == status2)
        self.assertFalse(status1 == status3)
        self.assertFalse(status1 == self.entity1)

    def test_status_and_string_equality_in_status_name(self):
        """testing a status can be compared with a string and returns True if
        the string matches the name and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertTrue(a_status == self.kwargs["name"])
        self.assertTrue(a_status == self.kwargs["name"].lower())
        self.assertTrue(a_status == self.kwargs["name"].upper())
        self.assertTrue(a_status == unicode(self.kwargs["name"]))
        self.assertTrue(a_status == unicode(self.kwargs["name"].lower()))
        self.assertTrue(a_status == unicode(self.kwargs["name"].upper()))
        self.assertFalse(a_status == "another name")
        self.assertFalse(a_status == u"another name")

    def test_status_and_string_equality_in_status_code(self):
        """testing a status can be compared with a string and returns True if
        the string matches the code and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertTrue(a_status == self.kwargs["code"])
        self.assertTrue(a_status == self.kwargs["code"].lower())
        self.assertTrue(a_status == self.kwargs["code"].upper())
        self.assertTrue(a_status == unicode(self.kwargs["code"]))
        self.assertTrue(a_status == unicode(self.kwargs["code"].lower()))
        self.assertTrue(a_status == unicode(self.kwargs["code"].upper()))

    def test_inequality(self):
        """testing inequality of two statuses
        """
        status1 = Status(**self.kwargs)
        status2 = Status(**self.kwargs)

        self.kwargs["name"] = "Work In Progress"
        self.kwargs["description"] = "use this when the object is still in \
        progress"
        self.kwargs["code"] = "WIP"

        status3 = Status(**self.kwargs)

        self.assertFalse(status1 != status2)
        self.assertTrue(status1 != status3)
        self.assertTrue(status1 != self.entity1)

    def test_status_and_string_inequality_in_status_name(self):
        """testing a status can be compared with a string and returns False if
        the string matches the name and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertFalse(a_status != self.kwargs["name"])
        self.assertFalse(a_status != self.kwargs["name"].lower())
        self.assertFalse(a_status != self.kwargs["name"].upper())
        self.assertFalse(a_status != unicode(self.kwargs["name"]))
        self.assertFalse(a_status != unicode(self.kwargs["name"].lower()))
        self.assertFalse(a_status != unicode(self.kwargs["name"].upper()))
        self.assertTrue(a_status != "another name")
        self.assertTrue(a_status != u"another name")

    def test_status_and_string_inequality_in_status_code(self):
        """testing a status can be compared with a string and returns False if
        the string matches the code and vice versa
        """
        a_status = Status(**self.kwargs)
        self.assertFalse(a_status != self.kwargs["code"])
        self.assertFalse(a_status != self.kwargs["code"].lower())
        self.assertFalse(a_status != self.kwargs["code"].upper())
        self.assertFalse(a_status != unicode(self.kwargs["code"]))
        self.assertFalse(a_status != unicode(self.kwargs["code"].lower()))
        self.assertFalse(a_status != unicode(self.kwargs["code"].upper()))
    
    def test_bg_color_argument_is_skipped(self):
        """testing if the bg_color attribute will be filled with the default
        color when the bg_color argument is skipped
        """
        self.kwargs.pop('bg_color')
        new_status = Status(**self.kwargs)
        self.assertEqual(
            defaults.status_bg_color,
            new_status.bg_color
        )
    
    def test_bg_color_argument_is_None(self):
        """testing if the bg_color attribute will be filled with the default
        color when the bg_color argument is None
        """
        self.kwargs['bg_color'] = None
        new_status = Status(**self.kwargs)
        self.assertEqual(
            defaults.status_bg_color,
            new_status.bg_color
        )
    
    def test_bg_color_attribute_is_None(self):
        """testing if the bg_color attribute is set to None will reset the
        color to default color
        """
        new_status = Status(**self.kwargs)
        new_status.bg_color = None
        self.assertEqual(
            defaults.status_bg_color,
            new_status.bg_color
        )
    
    def test_bg_color_argument_is_not_an_integer(self):
        """testing if a TypeError will be raised when the bg_color argument is
        not an integer
        """
        self.kwargs['bg_color'] = ["2342", 53, 33 ]
        self.assertRaises(TypeError, Status, **self.kwargs)
    
    def test_bg_color_attribute_is_not_an_integer(self):
        """testing if a TypeError will be raised when the bg_color attribute is
        not an integer
        """
        new_status = Status(**self.kwargs)
        self.assertRaises(TypeError, setattr, new_status, 'bg_color',
            ["asdf", 65, 25])
    
    def test_bg_color_argument_is_negative(self):
        """testing if the bg_color attribute value is clamped to 0 if bg_color
        argument is negative
        """
        self.kwargs['bg_color'] = -10
        new_status = Status(**self.kwargs)
        self.assertEqual(new_status.bg_color, 0)
    
    def test_bg_color_attribute_is_negative(self):
        """testing if the bg_color attribute value is clamped to 0 if it is set
        to a negative value
        """
        new_status = Status(**self.kwargs)
        new_status.bg_color = -10
        self.assertEqual(new_status.bg_color, 0)
    
    def test_bg_color_argument_is_too_big(self):
        """testing if the bg_color attribute value is clamped to 0xffffff if
        bg_color argument is bigger than that value
        """
        self.kwargs['bg_color'] = 0xffffff + 10
        new_status = Status(**self.kwargs)
        self.assertEqual(new_status.bg_color, 0xffffff)
    
    def test_bg_color_attribute_is_too_big(self):
        """testing if the bg_color attribute value is clamped to 0xffffff if it
        is bigger than that value
        """
        new_status = Status(**self.kwargs)
        new_status.bg_color = 0xffffff + 10
        self.assertEqual(new_status.bg_color, 0xffffff)
    
    def test_bg_color_argument_is_working_properly(self):
        """testing if the bg_color argument is working properly by setting its
        value its value to the bg_color attribute
        """
        self.kwargs['bg_color'] = 0xa0a000
        new_status = Status(**self.kwargs)
        self.assertEqual(self.kwargs['bg_color'], new_status.bg_color)
    
    def test_bg_color_attribute_is_working_properly(self):
        """testing if the bg_color attribute is working properly
        """
        new_status = Status(**self.kwargs)
        test_color = 0x564852
        new_status.bg_color = test_color
        self.assertEqual(test_color, new_status.bg_color)
    
    def test_fg_color_argument_is_skipped(self):
        """testing if the fg_color attribute will be filled with the default
        color when the fg_color argument is skipped
        """
        self.kwargs.pop('fg_color')
        new_status = Status(**self.kwargs)
        self.assertEqual(
            defaults.status_fg_color,
            new_status.fg_color
        )
    
    def test_fg_color_argument_is_None(self):
        """testing if the fg_color attribute will be filled with the default
        color when the fg_color argument is None
        """
        self.kwargs['fg_color'] = None
        new_status = Status(**self.kwargs)
        self.assertEqual(
            defaults.status_fg_color,
            new_status.fg_color
        )
    
    def test_fg_color_attribute_is_None(self):
        """testing if the fg_color attribute is set to None will reset the
        color to default color
        """
        new_status = Status(**self.kwargs)
        new_status.fg_color = None
        self.assertEqual(
            defaults.status_fg_color,
            new_status.fg_color
        )
    
    def test_fg_color_argument_is_not_an_integer(self):
        """testing if a TypeError will be raised when the fg_color argument is
        not an integer
        """
        self.kwargs['fg_color'] = ["2342", 53, 33 ]
        self.assertRaises(TypeError, Status, **self.kwargs)
    
    def test_fg_color_attribute_is_not_an_integer(self):
        """testing if a TypeError will be raised when the fg_color attribute is
        not an integer
        """
        new_status = Status(**self.kwargs)
        self.assertRaises(TypeError, setattr, new_status, 'fg_color',
            ["asdf", 65, 25])
    
    def test_fg_color_argument_is_negative(self):
        """testing if the fg_color attribute value is clamped to 0 if fg_color
        argument is negative
        """
        self.kwargs['fg_color'] = -10
        new_status = Status(**self.kwargs)
        self.assertEqual(new_status.fg_color, 0)
    
    def test_fg_color_attribute_is_negative(self):
        """testing if the fg_color attribute value is clamped to 0 if it is set
        to a negative value
        """
        new_status = Status(**self.kwargs)
        new_status.fg_color = -10
        self.assertEqual(new_status.fg_color, 0)
    
    def test_fg_color_argument_is_too_big(self):
        """testing if the fg_color attribute value is clamped to 0xffffff if
        fg_color argument is bigger than that value
        """
        self.kwargs['fg_color'] = 0xffffff + 10
        new_status = Status(**self.kwargs)
        self.assertEqual(new_status.fg_color, 0xffffff)
    
    def test_fg_color_attribute_is_too_big(self):
        """testing if the fg_color attribute value is clamped to 0xffffff if it
        is bigger than that value
        """
        new_status = Status(**self.kwargs)
        new_status.fg_color = 0xffffff + 10
        self.assertEqual(new_status.fg_color, 0xffffff)
    
    def test_fg_color_argument_is_working_properly(self):
        """testing if the fg_color argument is working properly by setting its
        value its value to the fg_color attribute
        """
        self.kwargs['fg_color'] = 0xa0a000
        new_status = Status(**self.kwargs)
        self.assertEqual(self.kwargs['fg_color'], new_status.fg_color)
    
    def test_fg_color_attribute_is_working_properly(self):
        """testing if the fg_color attribute is working properly
        """
        new_status = Status(**self.kwargs)
        test_color = 0x564852
        new_status.fg_color = test_color
        self.assertEqual(test_color, new_status.fg_color)
    



"""Version.version_of renamed to Version.task

Revision ID: 5355b569237b
Revises: 6297277da38
Create Date: 2013-06-10 11:47:28.984222

"""

# revision identifiers, used by Alembic.
revision = '5355b569237b'
down_revision = '6297277da38'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.alter_column('Versions', 'version_of_id', new_column_name='task_id')
    except sa.exc.OperationalError:
        # SQLite3
        # just create the new column
        # and copy data
        op.add_column(
            'Versions',
            'task_id',
            sa.Column(sa.Integer, sa.ForeignKey("Tasks.id"), nullable=False)
        )
        # copy data from Links.path to Links_Temp.full_path
        op.execute(
            'INSERT INTO "Versions".task_id '
            'SELECT "Versions".version_of_id '
            'FROM "Versions"'
        )

def downgrade():
    try:
        op.alter_column('Versions', 'task_id', new_column_name='version_of_id')
    except sa.exc.OperationalError:
        # SQLite3
        # just create the new column
        # and copy data
        op.add_column(
            'Versions',
            'version_of_id',
            sa.Column(sa.Integer, sa.ForeignKey("Tasks.id"), nullable=False)
        )
        op.execute(
            'INSERT INTO "Versions".version_of_id '
            'SELECT "Versions".task_id '
            'FROM "Versions"'
        )



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest2
import datetime

from stalker import (Entity, Link, Project, Repository, Status,
                     StatusList, Type, User, Scene)

class SceneTester(unittest2.TestCase):
    """Tests stalker.models.scene.Scene class
    """
    
    def setUp(self):
        """setup the test
        """
        # create statuses
        self.test_status1 = Status(name="Status1", code="STS1")
        self.test_status2 = Status(name="Status2", code="STS2")
        self.test_status3 = Status(name="Status3", code="STS3")
        self.test_status4 = Status(name="Status4", code="STS4")
        self.test_status5 = Status(name="Status5", code="STS5")

        # create a test project, user and a couple of shots
        self.project_type = Type(
            name="Test Project Type",
            code='test',
            target_entity_type=Project,
        )

        # create a status list for project
        self.project_status_list = StatusList(
            name="Project Statuses",
            statuses=[
                self.test_status1,
                self.test_status2,
                self.test_status3,
                self.test_status4,
                self.test_status5,
            ],
            target_entity_type=Project,
        )

        # create a repository
        self.repository_type = Type(
            name="Test Type",
            code='test',
            target_entity_type=Repository
        )

        self.test_repository = Repository(
            name="Test Repository",
            type=self.repository_type,
        )

        # create projects
        self.test_project = Project(
            name="Test Project 1",
            code='tp1',
            type=self.project_type,
            status_list=self.project_status_list,
            repository=self.test_repository,
        )

        self.test_project2 = Project(
            name="Test Project 2",
            code='tp2',
            type=self.project_type,
            status_list=self.project_status_list,
            repository=self.test_repository,
        )

        # the parameters
        self.kwargs = {
            "name": "Test Scene",
            'code': 'tsce',
            "description": "A test scene",
            "project": self.test_project,
        }

        # the test seuqence
        self.test_scene = Scene(**self.kwargs)
    
    def test___auto_name__class_attribute_is_set_to_False(self):
        """testing if the __auto_name__ class attribute is set to False for
        Scene class
        """
        self.assertFalse(Scene.__auto_name__)
    
    def test_shots_attribute_defaults_to_empty_list(self):
        """testing if the shots attribute defaults to an empty list
        """
        new_scene = Scene(**self.kwargs)
        self.assertEqual(new_scene.shots, [])

    def test_shots_attribute_is_set_None(self):
        """testing if a TypeError will be raised when the shots attribute will
        be set to None
        """
        self.assertRaises(TypeError, self.test_scene, "shots", None)

    def test_shots_attribute_is_set_to_other_than_a_list(self):
        """testing if a TypeError will be raised when the shots attribute is
        tried to be set to something other than a list
        """
        test_values = [1, 1.2, "a string"]
        for test_value in test_values:
            self.assertRaises(
                TypeError,
                setattr,
                self.test_scene,
                "shots",
                test_value
            )

    def test_shots_attribute_is_a_list_of_other_objects(self):
        """testing if a TypeError will be raised when the shots argument is a
        list of other type of objects
        """
        test_value = [1, 1.2, "a string"]
        self.assertRaises(
            TypeError,
            setattr,
            self.test_scene,
            "shots",
            test_value
        )

    def test_shots_attribute_elements_tried_to_be_set_to_non_Shot_object(self):
        """testing if a TypeError will be raised when the individual elements
        in the shots list tried to be set to something other than a Shot
        instance
        """
        test_values = [1, 1.2, "a string", ["a", "list"]]
        for test_value in test_values:
            self.assertRaises(TypeError,
                              self.test_scene.shots.append, test_value)

    def test_equality(self):
        """testing the equality of scenes
        """
        new_seq1 = Scene(**self.kwargs)
        new_seq2 = Scene(**self.kwargs)
        new_entity = Entity(**self.kwargs)

        self.kwargs["name"] = "a different scene"
        new_seq3 = Scene(**self.kwargs)

        self.assertTrue(new_seq1 == new_seq2)
        self.assertFalse(new_seq1 == new_seq3)
        self.assertFalse(new_seq1 == new_entity)

    def test_inequality(self):
        """testing the inequality of scenes
        """
        new_seq1 = Scene(**self.kwargs)
        new_seq2 = Scene(**self.kwargs)
        new_entity = Entity(**self.kwargs)

        self.kwargs["name"] = "a different scene"
        new_seq3 = Scene(**self.kwargs)

        self.assertFalse(new_seq1 != new_seq2)
        self.assertTrue(new_seq1 != new_seq3)
        self.assertTrue(new_seq1 != new_entity)

    def test_ProjectMixin_initialization(self):
        """testing if the ProjectMixin part is initialized correctly
        """
        status1 = Status(name="On Hold", code="OH")

        project_status_list = StatusList(
            name="Project Statuses", statuses=[status1],
            target_entity_type=Project
        )

        project_type = Type(
            name="Commercial",
            code='comm',
            target_entity_type=Project
        )

        new_project = Project(
            name="Test Project",
            code='tp',
            status=project_status_list[0],
            status_list=project_status_list,
            type=project_type,
            repository=self.test_repository
        )

        self.kwargs["project"] = new_project
        new_scene = Scene(**self.kwargs)
        self.assertEqual(new_scene.project, new_project)
    
    def test___strictly_typed___is_False(self):
        """testing if the __strictly_typed__ class attribute is False for
        Scene class.
        """
        self.assertEqual(Scene.__strictly_typed__, False)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest2

from sqlalchemy import Column, Integer, ForeignKey
from stalker.db.session import DBSession
from stalker.models.entity import SimpleEntity
from stalker.models.mixins import ReferenceMixin
from stalker.models.link import  Link

class DeclRefMixA(SimpleEntity, ReferenceMixin):
    """A test class for ReferenceMixin
    """

    __tablename__ = "DeclRefMixAs"
    __mapper_args__ = {"polymorphic_identity": "DeclRefMixA"}
    a_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                  primary_key=True)


    def __init__(self, **kwargs):
        super(DeclRefMixA, self).__init__(**kwargs)
        ReferenceMixin.__init__(self, **kwargs)


class DeclRefMixB(SimpleEntity, ReferenceMixin):
    """A test class for ReferenceMixin
    """

    __tablename__ = "RefMixBs"
    __mapper_args__ = {"polymorphic_identity": "DeclRefMixB"}
    b_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                  primary_key=True)


    def __init__(self, **kwargs):
        super(DeclRefMixB, self).__init__(**kwargs)
        ReferenceMixin.__init__(self, **kwargs)


class ReferenceMixinTester(unittest2.TestCase):
    """Tests ReferenceMixin
    """


    def test_ReferenceMixin_setup(self):
        """
        """
        a_ins = DeclRefMixA(name="ozgur")
        b_ins = DeclRefMixB(name="bozgur")

        new_link1 = Link(name="test link 1", full_path="none")
        new_link2 = Link(name="test link 2", full_path="no path")

        a_ins.references.append(new_link1)
        b_ins.references.append(new_link2)

        self.assertTrue(new_link1 in a_ins.references)
        self.assertTrue(new_link2 in b_ins.references)

    def tearDown(self):
        """clean up the test
        """
        if DBSession:
            DBSession.close()




